from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from dature.errors.exceptions import LineRange, SourceLocation
from dature.masking.masking import mask_env_line
from dature.types import NestedConflict, NestedConflicts

if TYPE_CHECKING:
    from dature.protocols import LoaderProtocol


@dataclass(frozen=True)
class ErrorContext:
    dataclass_name: str
    loader_class: "type[LoaderProtocol]"
    file_path: Path | None
    prefix: str | None
    split_symbols: str
    secret_paths: frozenset[str] = frozenset()
    mask_secrets: bool = False
    nested_conflicts: NestedConflicts | None = None


def read_file_content(file_path: Path | None) -> str | None:
    if file_path is None:
        return None

    with suppress(OSError):
        return file_path.read_text()

    return None


def _build_search_path(field_path: list[str], prefix: str | None) -> list[str]:
    if not prefix:
        return field_path
    prefix_parts = prefix.split(".")
    return prefix_parts + field_path


def _ranges_overlap(a: LineRange, b: LineRange) -> bool:
    return a.start <= b.end and b.start <= a.end


def _secret_overlaps_lines(
    *,
    file_content: str,
    line_range: LineRange,
    secret_paths: frozenset[str],
    prefix: str | None,
    path_finder_class: type,
) -> bool:
    finder = path_finder_class(file_content)
    for secret_path in secret_paths:
        search_path = _build_search_path(secret_path.split("."), prefix)
        secret_range = finder.find_line_range(search_path)
        if secret_range is not None and _ranges_overlap(line_range, secret_range):
            return True
    return False


def _resolve_conflict(
    field_path: list[str],
    ctx: ErrorContext,
) -> NestedConflict | None:
    if ctx.nested_conflicts is None:
        return None
    field_key = field_path[0] if field_path else ""
    return ctx.nested_conflicts.get(field_key)


def _apply_masking(
    locations: list[SourceLocation],
    ctx: ErrorContext,
    file_content: str | None,
    *,
    is_secret: bool,
) -> list[SourceLocation]:
    result: list[SourceLocation] = []
    for location in locations:
        should_mask = is_secret
        if (
            not should_mask
            and ctx.secret_paths
            and location.line_range is not None
            and ctx.loader_class.path_finder_class is not None
            and file_content is not None
        ):
            should_mask = _secret_overlaps_lines(
                file_content=file_content,
                line_range=location.line_range,
                secret_paths=ctx.secret_paths,
                prefix=ctx.prefix,
                path_finder_class=ctx.loader_class.path_finder_class,
            )
        if should_mask and location.line_content is not None:
            masked_lines = [mask_env_line(line) for line in location.line_content]
            result.append(
                SourceLocation(
                    display_label=location.display_label,
                    file_path=location.file_path,
                    line_range=location.line_range,
                    line_content=masked_lines,
                    env_var_name=location.env_var_name,
                ),
            )
        else:
            result.append(location)
    return result


def resolve_source_location(
    field_path: list[str],
    ctx: ErrorContext,
    file_content: str | None,
) -> list[SourceLocation]:
    is_secret = ".".join(field_path) in ctx.secret_paths
    conflict = _resolve_conflict(field_path, ctx)

    locations = ctx.loader_class.resolve_location(
        field_path,
        ctx.file_path,
        file_content,
        ctx.prefix,
        ctx.split_symbols,
        conflict,
    )

    return _apply_masking(locations, ctx, file_content, is_secret=is_secret)



"""Base64UrlBytes / Base64UrlStr — decoded from Base64."""

from base64 import urlsafe_b64decode
from dataclasses import dataclass

from dature.types import Base64UrlBytes, Base64UrlStr

encoded = "aGVsbG8gd29ybGQ="


@dataclass
class Config:
    token: Base64UrlStr
    data: Base64UrlBytes


config = Config(
    token=urlsafe_b64decode(encoded).decode(),
    data=urlsafe_b64decode(encoded),
)

assert config.token == "hello world"
assert type(config.token) is str
assert config.data == b"hello world"
assert type(config.data) is bytes



"""Tests for env_ module (EnvLoader and EnvFileLoader)."""

from dataclasses import dataclass
from pathlib import Path

import pytest

from dature import LoadMetadata, load
from dature.sources_loader.env_ import EnvFileLoader, EnvLoader
from examples.all_types_dataclass import EXPECTED_ALL_TYPES, AllPythonTypesCompact
from tests.sources_loader.checker import assert_all_types_equal


class TestEnvFileLoader:
    """Tests for EnvFileLoader class."""

    def test_prefix_filtering(self, prefixed_env_file: Path):
        """Test prefix filtering with nested structures."""
        loader = EnvFileLoader(prefix="APP_")
        result = loader.load_raw(prefixed_env_file)

        assert result.data == {
            "name": "PrefixedApp",
            "environment": "production",
            "database": {
                "host": "prod.db.com",
                "port": 5432,
            },
        }

    def test_custom_split_symbols(self, custom_separator_env_file: Path):
        """Test custom separator for nested keys."""
        loader = EnvFileLoader(prefix="APP_", split_symbols=".")
        result = loader.load_raw(custom_separator_env_file)

        assert result.data == {
            "name": "CustomApp",
            "environment": "development",
            "db": {
                "host": "dev.db.com",
                "port": 5432,
            },
        }

    def test_comprehensive_type_conversion(self, all_types_env_file: Path):
        """Test loading ENV with full type coercion to dataclass."""
        result = load(LoadMetadata(file_=all_types_env_file, loader=EnvFileLoader), AllPythonTypesCompact)

        assert_all_types_equal(result, EXPECTED_ALL_TYPES)

    def test_empty_file(self, tmp_path: Path):
        """Test loading empty .env file."""
        env_file = tmp_path / ".env"
        env_file.write_text("")

        loader = EnvFileLoader()
        data = loader._load(env_file)

        assert data == {}

    def test_env_file_env_var_substitution(self, tmp_path: Path, monkeypatch):
        monkeypatch.setenv("BASE_URL", "https://api.example.com")

        env_file = tmp_path / ".env"
        env_file.write_text("api_url=$BASE_URL/v1\nbase=$BASE_URL")

        @dataclass
        class Config:
            api_url: str
            base: str

        result = load(LoadMetadata(file_=env_file, loader=EnvFileLoader), Config)

        assert result.api_url == "https://api.example.com/v1"
        assert result.base == "https://api.example.com"

    def test_env_file_env_var_partial_substitution(self, tmp_path: Path, monkeypatch):
        monkeypatch.setenv("HOST", "localhost")
        monkeypatch.setenv("PORT", "8080")

        env_file = tmp_path / ".env"
        env_file.write_text("url=http://${HOST}:${PORT}/api")

        @dataclass
        class Config:
            url: str

        result = load(LoadMetadata(file_=env_file, loader=EnvFileLoader), Config)

        assert result.url == "http://localhost:8080/api"

    def test_env_file_dollar_sign_mid_string_existing_var(self, tmp_path: Path, monkeypatch):
        monkeypatch.setenv("abc", "replaced")

        env_file = tmp_path / ".env"
        env_file.write_text("value=prefix$abc/suffix")

        @dataclass
        class Config:
            value: str

        result = load(LoadMetadata(file_=env_file, loader=EnvFileLoader), Config)

        assert result.value == "prefixreplaced/suffix"

    @pytest.mark.parametrize(
        ("raw_value", "expected"),
        [
            ('"hello"', "hello"),
            ("'hello'", "hello"),
            ('""hello""', '"hello"'),
            ("''hello''", "'hello'"),
            ("\"'hello'\"", "'hello'"),
            ("'\"hello\"'", '"hello"'),
            ("\"hello'", "\"hello'"),
            ("'hello\"", "'hello\""),
            ("hello", "hello"),
            ('""', ""),
            ("''", ""),
            ('"', '"'),
            ("'", "'"),
        ],
    )
    def test_quote_stripping(self, tmp_path: Path, raw_value: str, expected: str):
        env_file = tmp_path / ".env"
        env_file.write_text(f"value={raw_value}")

        loader = EnvFileLoader()
        data = loader._load(env_file)

        assert data == {"value": expected}

    def test_env_file_dollar_sign_mid_string_missing_var(self, tmp_path: Path, monkeypatch):
        monkeypatch.delenv("nonexistent", raising=False)

        env_file = tmp_path / ".env"
        env_file.write_text("value=prefix$nonexistent/suffix")

        @dataclass
        class Config:
            value: str

        result = load(LoadMetadata(file_=env_file, loader=EnvFileLoader), Config)

        assert result.value == "prefix$nonexistent/suffix"


class TestEnvLoader:
    """Tests for EnvLoader class."""

    def test_comprehensive_type_conversion(self, monkeypatch):
        """Test loading from os.environ with full type coercion to dataclass."""
        env_vars = {
            # Scalars
            "APP_STRING_VALUE": "hello world",
            "APP_INTEGER_VALUE": "42",
            "APP_FLOAT_VALUE": "3.14159",
            "APP_BOOLEAN_VALUE": "true",
            "APP_NONE_VALUE": "",
            # Numeric
            "APP_DECIMAL_VALUE": "3.14159265358979323846264338327950288",
            "APP_FLOAT_INF": "inf",
            "APP_FLOAT_NAN": "nan",
            # Date/time
            "APP_DATE_VALUE": "2024-01-15",
            "APP_DATETIME_VALUE": "2024-01-15T10:30:00",
            "APP_DATETIME_VALUE_WITH_TIMEZONE": "2024-01-15T10:30:00+03:00",
            "APP_DATETIME_VALUE_WITH_Z_TIMEZONE": "2024-01-15T10:30:00Z",
            "APP_TIME_VALUE": "10:30:00",
            "APP_TIMEDELTA_VALUE_WITH_DAY": "1 day, 2:30:00",
            "APP_TIMEDELTA_VALUE_WITHOUT_DAY": "2:30:00",
            "APP_TIMEDELTA_VALUE_WITHOUT_SECONDS": "2:30",
            # Lists
            "APP_LIST_STRINGS": '["item1","item2","item3"]',
            "APP_LIST_INTEGERS": "[1,2,3,4,5]",
            "APP_LIST_MIXED": '["string",42,3.14,true,null]',
            "APP_LIST_NESTED": '[["a","b"],["c","d"]]',
            "APP_LIST_DICTS": '[{"name":"Alice","age":30},{"name":"Bob","age":25}]',
            # Tuples
            "APP_TUPLE_SIMPLE": "[1,2,3]",
            "APP_TUPLE_MIXED": '["text",42,true]',
            "APP_TUPLE_NESTED": '[[1,2,3],["a","b","c"]]',
            # Sets
            "APP_SET_INTEGERS": "[1,2,3,4,5]",
            "APP_SET_STRINGS": '["apple","banana","cherry"]',
            # Dicts
            "APP_DICT_SIMPLE__KEY1": "value1",
            "APP_DICT_SIMPLE__KEY2": "value2",
            "APP_DICT_MIXED__STRING": "text",
            "APP_DICT_MIXED__NUMBER": "42",
            "APP_DICT_MIXED__FLOAT": "3.14",
            "APP_DICT_MIXED__BOOL": "true",
            "APP_DICT_MIXED__LIST": "[1,2,3]",
            "APP_DICT_NESTED__LEVEL1__LEVEL2__LEVEL3": "deep_value",
            "APP_DICT_INT_KEYS__1": "one",
            "APP_DICT_INT_KEYS__2": "two",
            "APP_DICT_INT_KEYS__3": "three",
            "APP_DICT_LIST_DICT": (
                '{"users":[{"name":"Alice","role":"admin"},{"name":"Bob","role":"user"}],'
                '"teams":[{"name":"backend","size":5}]}'
            ),
            # Binary/encoding
            "APP_BYTES_VALUE": "binary data",
            "APP_BYTEARRAY_VALUE": "binary",
            "APP_COMPLEX_VALUE": "1+2j",
            "APP_BASE64URL_BYTES_VALUE": "SGVsbG8gV29ybGQ=",
            "APP_BASE64URL_STR_VALUE": "c2VjcmV0IHRva2Vu",
            # Custom fields
            "APP_SECRET_STR_VALUE": "supersecret123",
            "APP_PAYMENT_CARD_NUMBER_VALUE": "4111111111111111",
            "APP_BYTE_SIZE_VALUE": "1.5 GB",
            # Paths
            "APP_PATH_VALUE": "/usr/local/bin",
            "APP_PURE_POSIX_PATH_VALUE": "/etc/hosts",
            "APP_PURE_WINDOWS_PATH_VALUE": "C:/Windows/System32",
            # Network
            "APP_IPV4_ADDRESS_VALUE": "192.168.1.1",
            "APP_IPV6_ADDRESS_VALUE": "2001:db8::1",
            "APP_IPV4_NETWORK_VALUE": "192.168.1.0/24",
            "APP_IPV6_NETWORK_VALUE": "2001:db8::/32",
            "APP_IPV4_INTERFACE_VALUE": "192.168.1.1/24",
            "APP_IPV6_INTERFACE_VALUE": "2001:db8::1/32",
            # Identifiers
            "APP_UUID_VALUE": "550e8400-e29b-41d4-a716-446655440000",
            "APP_URL_VALUE": "https://example.com/path?query=value#fragment",
            # Edge cases
            "APP_EMPTY_STRING": "",
            "APP_EMPTY_LIST": "[]",
            "APP_EMPTY_DICT": "{}",
            "APP_ZERO_INT": "0",
            "APP_ZERO_FLOAT": "0.0",
            "APP_FALSE_BOOL": "false",
            # Union/optional
            "APP_OPTIONAL_STRING": "",
            "APP_UNION_TYPE": "42",
            "APP_NESTED_OPTIONAL": '[{"name":"Alice","email":"alice@example.com"},{"name":"Bob","email":null}]',
            "APP_RANGE_VALUES": "[0,2,4,6,8]",
            "APP_FROZENSET_VALUE": "[1,2,3,4,5]",
            # Nested dataclasses in collections
            "APP_NESTED_DC_SINGLE__CITY": "Moscow",
            "APP_NESTED_DC_SINGLE__ZIP_CODE": "101000",
            "APP_NESTED_DC_LIST": '[{"name":"urgent","priority":1},{"name":"low","priority":5}]',
            "APP_NESTED_DC_DICT": (
                '{"home":{"city":"Berlin","zip_code":"10115"},"work":{"city":"Paris","zip_code":"75001"}}'
            ),
            "APP_NESTED_DC_TUPLE": '[{"name":"bug","priority":2},{"name":"feature","priority":3}]',
            # Enum/Flag/Literal
            "APP_ENUM_VALUE": "green",
            "APP_FLAG_VALUE": "3",
            "APP_LITERAL_VALUE": "info",
            # Additional stdlib types
            "APP_REGEX_PATTERN": "^[a-z]+$",
            "APP_FRACTION_VALUE": "1/3",
            "APP_DEQUE_VALUE": '["first","second","third"]',
        }
        for key, value in env_vars.items():
            monkeypatch.setenv(key, value)

        result = load(LoadMetadata(loader=EnvLoader, prefix="APP_"), AllPythonTypesCompact)

        assert_all_types_equal(result, EXPECTED_ALL_TYPES)

    def test_prefix_filtering(self, monkeypatch):
        """Test that only variables with prefix are loaded."""
        monkeypatch.setenv("APP_VAR", "included")
        monkeypatch.setenv("OTHER_VAR", "excluded")
        monkeypatch.setenv("APP_KEY", "also_included")

        @dataclass
        class TestConfig:
            var: str
            key: str

        expected_data = TestConfig(var="included", key="also_included")

        data = load(LoadMetadata(loader=EnvLoader, prefix="APP_"), TestConfig)

        assert data == expected_data

    def test_custom_split_symbols(self, monkeypatch):
        """Test custom separator for nested keys."""
        monkeypatch.setenv("APP_DB.HOST", "localhost")
        monkeypatch.setenv("APP_DB.PORT", "5432")

        @dataclass
        class TestData:
            host: str
            port: str

        @dataclass
        class TestConfig:
            db: TestData

        expected_data = TestConfig(db=TestData(host="localhost", port="5432"))

        data = load(
            LoadMetadata(loader=EnvLoader, prefix="APP_", split_symbols="."),
            TestConfig,
        )

        assert data == expected_data



"""Tests for FieldPath lazy field path builder."""

from dataclasses import dataclass

import pytest

from dature.field_path import F, FieldPath, validate_field_path_owner


@dataclass
class _Database:
    uri: str


@dataclass
class _Cfg:
    host: str
    database: _Database


class TestFieldPath:
    @pytest.mark.parametrize(
        ("field_path", "expected_path"),
        [
            pytest.param(F[_Cfg].host, "host", id="single_field_class"),
            pytest.param(F[_Cfg].database.uri, "database.uri", id="nested_field_class"),
            pytest.param(F["Config"].host, "host", id="single_field_string"),
            pytest.param(F["Config"].database.uri, "database.uri", id="nested_field_string"),
        ],
    )
    def test_path(self, field_path: FieldPath, expected_path: str):
        assert field_path.as_path() == expected_path

    def test_owner_is_class(self):
        assert F[_Cfg].host.owner is _Cfg

    def test_owner_is_string(self):
        assert F["Config"].host.owner == "Config"

    def test_no_fields_raises_value_error(self):
        with pytest.raises(ValueError, match="at least one field name"):
            F[_Cfg].as_path()

    def test_is_frozen(self):
        fp = F[_Cfg].host
        with pytest.raises(AttributeError):
            fp.owner = "other"

    def test_nonexistent_field_raises_attribute_error(self):
        with pytest.raises(AttributeError, match="'_Cfg' has no field 'nonexistent'"):
            _ = F[_Cfg].nonexistent

    def test_not_a_dataclass_raises_type_error(self):
        class Plain:
            pass

        with pytest.raises(TypeError, match="'Plain' is not a dataclass"):
            F[Plain]

    def test_string_owner_skips_validation(self):
        fp = F["Whatever"].anything.deep.path
        assert fp.as_path() == "anything.deep.path"


class TestValidateFieldPathOwner:
    def test_string_owner_matches(self):
        validate_field_path_owner(F["_Cfg"].host, _Cfg)

    def test_type_owner_matches(self):
        validate_field_path_owner(F[_Cfg].host, _Cfg)

    def test_string_owner_mismatch(self):
        with pytest.raises(TypeError) as exc_info:
            validate_field_path_owner(F["Other"].host, _Cfg)
        assert str(exc_info.value) == "FieldPath owner 'Other' does not match target dataclass '_Cfg'"

    def test_type_owner_mismatch(self):
        with pytest.raises(TypeError) as exc_info:
            validate_field_path_owner(F[_Database].uri, _Cfg)
        assert str(exc_info.value) == "FieldPath owner '_Database' does not match target dataclass '_Cfg'"

    def test_string_owner_nonexistent_field(self):
        with pytest.raises(AttributeError, match="'_Cfg' has no field 'nonexistent'"):
            validate_field_path_owner(F["_Cfg"].nonexistent, _Cfg)

    def test_string_owner_nonexistent_nested_field(self):
        with pytest.raises(AttributeError, match="'_Database' has no field 'missing'"):
            validate_field_path_owner(F["_Cfg"].database.missing, _Cfg)

    def test_string_owner_valid_nested_path(self):
        validate_field_path_owner(F["_Cfg"].database.uri, _Cfg)



"""Decorator mode — auto-load config from a YAML file."""

from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, load

SHARED_DIR = Path(__file__).parents[1] / "shared"


@load(LoadMetadata(file_=SHARED_DIR / "common_app.yaml"))
@dataclass
class Config:
    host: str
    port: int
    debug: bool = False


config = Config()

assert config.host == "localhost"
assert config.port == 8080
assert config.debug is False



"""Tests for skip_field_provider.py."""

from dataclasses import dataclass

from adaptix import Retort

from dature.skip_field_provider import (
    ModelToDictProvider,
    SkipFieldProvider,
    filter_invalid_fields,
)


class TestFilterInvalidFields:
    def test_all_fields_valid(self):
        @dataclass
        class Config:
            host: str
            port: int

        probe = Retort(
            strict_coercion=False,
            recipe=[SkipFieldProvider(), ModelToDictProvider()],
        )
        raw = {"host": "localhost", "port": 8080}
        result = filter_invalid_fields(raw, probe, Config, None)

        assert result.cleaned_dict == {"host": "localhost", "port": 8080}
        assert result.skipped_paths == []

    def test_one_field_invalid(self):
        @dataclass
        class Config:
            host: str
            port: int

        probe = Retort(
            strict_coercion=False,
            recipe=[SkipFieldProvider(), ModelToDictProvider()],
        )
        raw = {"host": "localhost", "port": "abc"}
        result = filter_invalid_fields(raw, probe, Config, None)

        assert result.cleaned_dict == {"host": "localhost"}
        assert result.skipped_paths == ["port"]

    def test_nested_field_invalid(self):
        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            db: Database

        probe = Retort(
            strict_coercion=False,
            recipe=[SkipFieldProvider(), ModelToDictProvider()],
        )
        raw = {"db": {"host": "localhost", "port": "abc"}}
        result = filter_invalid_fields(raw, probe, Config, None)

        assert result.cleaned_dict == {"db": {"host": "localhost"}}
        assert result.skipped_paths == ["db.port"]

    def test_allowed_fields_restricts_skip(self):
        @dataclass
        class Config:
            host: str
            port: int
            timeout: int

        probe = Retort(
            strict_coercion=False,
            recipe=[SkipFieldProvider(), ModelToDictProvider()],
        )
        raw = {"host": "localhost", "port": "abc", "timeout": "bad"}
        result = filter_invalid_fields(raw, probe, Config, {"port"})

        assert result.cleaned_dict == {"host": "localhost", "timeout": "bad"}
        assert result.skipped_paths == ["port"]

    def test_non_dict_input(self):
        @dataclass
        class Config:
            host: str

        probe = Retort(
            strict_coercion=False,
            recipe=[SkipFieldProvider(), ModelToDictProvider()],
        )
        result = filter_invalid_fields("not a dict", probe, Config, None)

        assert result.cleaned_dict == "not a dict"
        assert result.skipped_paths == []



"""Tests for yaml_ module (Yaml12Loader)."""

from dataclasses import dataclass
from pathlib import Path

import pytest

from dature import LoadMetadata, load
from dature.errors.exceptions import DatureConfigError, FieldLoadError
from dature.sources_loader.yaml_ import Yaml12Loader
from examples.all_types_dataclass import (
    EXPECTED_ALL_TYPES,
    AllPythonTypesCompact,
)
from tests.sources_loader.checker import assert_all_types_equal


class TestYaml12Loader:
    """Tests for Yaml12Loader class."""

    def test_comprehensive_type_conversion(self, all_types_yaml12_file: Path):
        """Test loading YAML with full type coercion to dataclass."""
        result = load(LoadMetadata(file_=all_types_yaml12_file, loader=Yaml12Loader), AllPythonTypesCompact)

        assert_all_types_equal(result, EXPECTED_ALL_TYPES)

    def test_yaml_with_prefix(self, prefixed_yaml_file: Path):
        @dataclass
        class PrefixedConfig:
            name: str
            port: int
            debug: bool
            environment: str

        expected_data = PrefixedConfig(
            name="PrefixedApp",
            port=9000,
            debug=False,
            environment="production",
        )

        result = load(
            LoadMetadata(file_=prefixed_yaml_file, loader=Yaml12Loader, prefix="app"),
            PrefixedConfig,
        )

        assert result == expected_data

    def test_yaml_env_var_substitution(self, yaml_config_with_env_vars_file: Path, monkeypatch):
        """Test YAML environment variable substitution."""
        monkeypatch.setenv("DATABASE_URL", "postgresql://localhost/db")
        monkeypatch.setenv("SECRET_KEY", "my_secret")
        monkeypatch.setenv("REDIS_HOST", "redis.local")
        monkeypatch.setenv("QUEUE_HOST", "queue.local")

        @dataclass
        class Services:
            cache: dict[str, str]
            queue: dict[str, str]

        @dataclass
        class EnvConfig:
            database_url: str
            secret_key: str
            services: Services

        result = load(
            LoadMetadata(file_=yaml_config_with_env_vars_file, loader=Yaml12Loader),
            EnvConfig,
        )

        assert result.database_url == "postgresql://localhost/db"
        assert result.secret_key == "my_secret"
        assert result.services.cache == {"host": "redis.local"}
        assert result.services.queue == {"host": "queue.local"}

    def test_yaml_env_var_partial_substitution(self, tmp_path: Path, monkeypatch):
        monkeypatch.setenv("HOST", "localhost")
        monkeypatch.setenv("PORT", "8080")

        yaml_file = tmp_path / "env.yaml"
        yaml_file.write_text('url: "http://${HOST}:${PORT}/api"')

        @dataclass
        class Config:
            url: str

        result = load(LoadMetadata(file_=yaml_file, loader=Yaml12Loader), Config)

        assert result.url == "http://localhost:8080/api"

    def test_yaml_dollar_sign_mid_string_existing_var(self, tmp_path: Path, monkeypatch):
        monkeypatch.setenv("abc", "replaced")

        yaml_file = tmp_path / "dollar.yaml"
        yaml_file.write_text("value: prefix$abc/suffix")

        @dataclass
        class Config:
            value: str

        result = load(LoadMetadata(file_=yaml_file, loader=Yaml12Loader), Config)

        assert result.value == "prefixreplaced/suffix"

    def test_yaml_dollar_sign_mid_string_missing_var(self, tmp_path: Path, monkeypatch):
        monkeypatch.delenv("nonexistent", raising=False)

        yaml_file = tmp_path / "dollar.yaml"
        yaml_file.write_text("value: prefix$nonexistent/suffix")

        @dataclass
        class Config:
            value: str

        result = load(LoadMetadata(file_=yaml_file, loader=Yaml12Loader), Config)

        assert result.value == "prefix$nonexistent/suffix"

    def test_yaml_empty_file(self, tmp_path: Path):
        """Test loading empty YAML file."""
        yaml_file = tmp_path / "empty.yaml"
        yaml_file.write_text("")

        loader = Yaml12Loader()
        data = loader._load(yaml_file)

        assert data is None

    def test_bool_in_int_field_raises_error(self, tmp_path: Path):
        yaml_file = tmp_path / "config.yaml"
        yaml_file.write_text("count: true")

        @dataclass
        class Config:
            count: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(LoadMetadata(file_=yaml_file, loader=Yaml12Loader), Config)

        err = exc_info.value
        assert len(err.exceptions) == 1
        first = err.exceptions[0]
        assert isinstance(first, FieldLoadError)
        assert first.field_path == ["count"]
        assert str(first) == (
            f"  [count]  Expected int, got bool: True\n   └── FILE '{yaml_file}', line 1\n       count: true"
        )

    def test_int_in_bool_field_raises_error(self, tmp_path: Path):
        yaml_file = tmp_path / "config.yaml"
        yaml_file.write_text("flag: 1")

        @dataclass
        class Config:
            flag: bool

        with pytest.raises(DatureConfigError) as exc_info:
            load(LoadMetadata(file_=yaml_file, loader=Yaml12Loader), Config)

        err = exc_info.value
        assert len(err.exceptions) == 1
        first = err.exceptions[0]
        assert isinstance(first, FieldLoadError)
        assert first.field_path == ["flag"]
        assert str(first) == (f"  [flag]  Expected bool, got int\n   └── FILE '{yaml_file}', line 1\n       flag: 1")



"""Tests for custom loaders — subclassing BaseLoader."""

import xml.etree.ElementTree as ET
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar

from adaptix import Provider, loader

from dature import LoadMetadata, load
from dature.sources_loader.base import BaseLoader
from dature.sources_loader.loaders import bool_loader, float_from_string
from dature.types import FileOrStream, JSONValue


class XmlLoader(BaseLoader):
    display_name: ClassVar[str] = "xml"

    def _load(self, path: FileOrStream) -> JSONValue:
        if not isinstance(path, Path):
            msg = "XmlLoader only supports file paths"
            raise TypeError(msg)
        tree = ET.parse(path)  # noqa: S314
        root = tree.getroot()
        return {child.tag: child.text or "" for child in root}

    def _additional_loaders(self) -> list[Provider]:
        return [
            loader(bool, bool_loader),
            loader(float, float_from_string),
        ]


@dataclass
class XmlConfig:
    host: str
    port: int
    debug: bool


class TestCustomLoader:
    def test_xml_loader(self, tmp_path: Path) -> None:
        xml_file = tmp_path / "config.xml"
        xml_file.write_text(
            "<config>\n    <host>localhost</host>\n    <port>9090</port>\n    <debug>true</debug>\n</config>\n",
        )

        result = load(
            LoadMetadata(file_=xml_file, loader=XmlLoader),
            XmlConfig,
        )

        assert result.host == "localhost"
        assert result.port == 9090
        assert result.debug is True



from dataclasses import dataclass
from pathlib import Path

import pytest

from dature import LoadMetadata, load
from dature.errors.exceptions import DatureConfigError
from dature.validators.root import RootValidator


class TestRootValidator:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            port: int
            user: str

        def validate_config(obj: Config) -> bool:
            if obj.port < 1024:
                return obj.user == "root"
            return True

        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 80, "user": "root"}')

        metadata = LoadMetadata(
            file_=json_file,
            root_validators=(RootValidator(func=validate_config),),
        )
        result = load(metadata, Config)

        assert result.port == 80
        assert result.user == "root"

    def test_validation_fails(self, tmp_path: Path):
        @dataclass
        class Config:
            port: int
            user: str

        def validate_config(obj: Config) -> bool:
            if obj.port < 1024:
                return obj.user == "root"
            return True

        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 80, "user": "admin"}')

        metadata = LoadMetadata(
            file_=json_file,
            root_validators=(RootValidator(func=validate_config),),
        )

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (f"  [<root>]  Root validation failed\n   └── FILE '{json_file}'")

    def test_multiple_root_validators(self, tmp_path: Path):
        @dataclass
        class Config:
            min_value: int
            max_value: int
            step: int

        def validate_min_max(obj: Config) -> bool:
            return obj.min_value < obj.max_value

        def validate_step(obj: Config) -> bool:
            return obj.step > 0

        json_file = tmp_path / "config.json"
        json_file.write_text('{"min_value": 10, "max_value": 100, "step": 5}')

        metadata = LoadMetadata(
            file_=json_file,
            root_validators=(
                RootValidator(func=validate_min_max),
                RootValidator(func=validate_step),
            ),
        )
        result = load(metadata, Config)

        assert result.min_value == 10
        assert result.max_value == 100
        assert result.step == 5

    def test_multiple_root_validators_all_fail(self, tmp_path: Path):
        @dataclass
        class Config:
            min_value: int
            max_value: int
            step: int

        def validate_min_max(obj: Config) -> bool:
            return obj.min_value < obj.max_value

        def validate_step(obj: Config) -> bool:
            return obj.step > 0

        json_file = tmp_path / "config.json"
        json_file.write_text('{"min_value": 100, "max_value": 10, "step": -5}')

        metadata = LoadMetadata(
            file_=json_file,
            root_validators=(
                RootValidator(func=validate_min_max),
                RootValidator(func=validate_step),
            ),
        )

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (f"  [<root>]  Root validation failed\n   └── FILE '{json_file}'")

    def test_root_validator_privileged_port(self, tmp_path: Path):
        @dataclass
        class Config:
            port: int
            host: str

        def validate_config(obj: Config) -> bool:
            return obj.host != "localhost" or obj.port in range(1024, 65536)

        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 80, "host": "localhost"}')

        metadata = LoadMetadata(
            file_=json_file,
            root_validators=(RootValidator(func=validate_config),),
        )

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (f"  [<root>]  Root validation failed\n   └── FILE '{json_file}'")

    def test_root_validator_with_decorator(self, tmp_path: Path):
        def validate_credentials(obj) -> bool:
            if obj.username == "admin":
                return len(obj.password) >= 12
            return len(obj.password) >= 8

        json_file = tmp_path / "config.json"
        json_file.write_text('{"username": "admin", "password": "short"}')

        metadata = LoadMetadata(
            file_=json_file,
            root_validators=(RootValidator(func=validate_credentials),),
        )

        @load(metadata)
        @dataclass
        class Credentials:
            username: str
            password: str

        with pytest.raises(DatureConfigError) as exc_info:
            Credentials()

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Credentials loading errors (1)"
        assert str(e.exceptions[0]) == (f"  [<root>]  Root validation failed\n   └── FILE '{json_file}'")

    def test_custom_error_message(self, tmp_path: Path):
        @dataclass
        class Config:
            port: int
            user: str

        def validate_config(obj: Config) -> bool:
            if obj.port < 1024:
                return obj.user == "root"
            return True

        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 80, "user": "admin"}')

        metadata = LoadMetadata(
            file_=json_file,
            root_validators=(
                RootValidator(
                    func=validate_config,
                    error_message="Ports below 1024 require root user",
                ),
            ),
        )

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (f"  [<root>]  Ports below 1024 require root user\n   └── FILE '{json_file}'")



"""Tests for skip_if_invalid / skip_invalid_fields feature."""

import logging
from dataclasses import dataclass
from pathlib import Path

import pytest

from dature import F, LoadMetadata, MergeMetadata, MergeStrategy, load
from dature.errors.exceptions import DatureConfigError


class TestMergeSkipInvalidFields:
    def test_fallback_to_other_source(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc"}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=source1),
                    LoadMetadata(file_=source2),
                ),
                skip_invalid_fields=True,
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080

    def test_all_sources_invalid_with_default(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc"}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"port": "def"}')

        @dataclass
        class Config:
            host: str
            port: int = 9090

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=source1),
                    LoadMetadata(file_=source2),
                ),
                skip_invalid_fields=True,
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 9090

    def test_all_sources_invalid_no_default_raises(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc"}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"port": "def"}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=source1),
                        LoadMetadata(file_=source2),
                    ),
                    skip_invalid_fields=True,
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == (
            f"  [port]  Missing required field (invalid in: json '{source1}', json '{source2}')\n"
            f"   ├── FILE '{source1}', line 1\n"
            f'       {{"host": "localhost", "port": "abc"}}\n'
            f"   └── FILE '{source2}', line 1\n"
            f'       {{"port": "def"}}'
        )

    def test_nested_dataclass(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"db": {"host": "s1-host", "port": "abc"}}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"db": {"host": "s2-host", "port": 5432}}')

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            db: Database

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=source1),
                    LoadMetadata(file_=source2),
                ),
                skip_invalid_fields=True,
            ),
            Config,
        )

        assert result.db.host == "s2-host"
        assert result.db.port == 5432

    def test_per_source_override(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc"}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=source1, skip_if_invalid=True),
                    LoadMetadata(file_=source2),
                ),
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080

    def test_global_flag(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc"}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=source1),
                    LoadMetadata(file_=source2),
                ),
                skip_invalid_fields=True,
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080

    def test_backward_compat_no_skip(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc"}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(LoadMetadata(file_=source1),),
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == (
            f"  [port]  invalid literal for int() with base 10: 'abc'\n"
            f"   └── FILE '{source1}', line 1\n"
            f'       {{"host": "localhost", "port": "abc"}}'
        )

    def test_raise_on_conflict_with_skip(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc"}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"host": "localhost", "port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=source1),
                    LoadMetadata(file_=source2),
                ),
                strategy=MergeStrategy.RAISE_ON_CONFLICT,
                skip_invalid_fields=True,
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080

    def test_skip_specific_fields_only(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc", "timeout": "bad"}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int
            timeout: int = 30

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(
                        file_=source1,
                        skip_if_invalid=(F[Config].port, F[Config].timeout),
                    ),
                    LoadMetadata(file_=source2),
                ),
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080
        assert result.timeout == 30

    def test_skip_specific_fields_non_listed_field_raises(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": 123, "port": "abc"}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(
                            file_=source1,
                            skip_if_invalid=(F[Config].port,),
                        ),
                    ),
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 2
        assert str(err) == "Config loading errors (2)"
        assert str(err.exceptions[0]) == (
            f'  [host]  Expected str, got int\n   └── FILE \'{source1}\', line 1\n       {{"host": 123, "port": "abc"}}'
        )
        assert str(err.exceptions[1]) == (
            f"  [port]  Missing required field (invalid in: json '{source1}')\n"
            f"   └── FILE '{source1}', line 1\n"
            f'       {{"host": 123, "port": "abc"}}'
        )

    def test_log_warnings(self, tmp_path: Path, caplog: pytest.LogCaptureFixture):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"host": "localhost", "port": "abc"}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        with caplog.at_level(logging.WARNING, logger="dature"):
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=source1),
                        LoadMetadata(file_=source2),
                    ),
                    skip_invalid_fields=True,
                ),
                Config,
            )

        warning_messages = [r.getMessage() for r in caplog.records if r.levelno >= logging.WARNING]
        assert warning_messages == [
            "[Config] Source 0: Skipped invalid field 'port'",
            "[Config] Source 1: Skipped invalid field 'host'",
        ]


class TestSingleSourceSkipInvalidFields:
    def test_skip_with_default(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": "abc"}')

        @dataclass
        class Config:
            host: str
            port: int = 8080

        result = load(
            LoadMetadata(file_=json_file, skip_if_invalid=True),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080

    def test_skip_without_default_raises(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": "abc"}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                LoadMetadata(file_=json_file, skip_if_invalid=True),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == (
            f"  [port]  Missing required field (invalid in: json '{json_file}')\n"
            f"   └── FILE '{json_file}', line 1\n"
            f'       {{"host": "localhost", "port": "abc"}}'
        )

    def test_single_source_decorator_skip(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": "abc"}')

        @load(LoadMetadata(file_=json_file, skip_if_invalid=True))
        @dataclass
        class Config:
            host: str
            port: int = 8080

        cfg = Config()
        assert cfg.host == "localhost"
        assert cfg.port == 8080

    def test_single_source_specific_fields(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": "abc", "timeout": 60}')

        @dataclass
        class Config:
            host: str
            port: int = 9090
            timeout: int = 30

        result = load(
            LoadMetadata(
                file_=json_file,
                skip_if_invalid=(F[Config].port,),
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 9090
        assert result.timeout == 60

    def test_single_source_log_warnings(self, tmp_path: Path, caplog: pytest.LogCaptureFixture):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": "abc"}')

        @dataclass
        class Config:
            host: str
            port: int = 8080

        with caplog.at_level(logging.WARNING, logger="dature"):
            load(
                LoadMetadata(file_=json_file, skip_if_invalid=True),
                Config,
            )

        warning_messages = [r.getMessage() for r in caplog.records if r.levelno >= logging.WARNING]
        assert warning_messages == ["[Config] Skipped invalid field 'port'"]


class TestSkipInvalidSameFieldNameNested:
    def test_skip_root_field_not_nested(self, tmp_path: Path):
        source = tmp_path / "config.json"
        source.write_text('{"port": "abc", "inner": {"port": 9090}}')

        @dataclass
        class Inner:
            port: int

        @dataclass
        class Config:
            port: int = 3000
            inner: Inner = None  # type: ignore[assignment]

        result = load(
            LoadMetadata(
                file_=source,
                skip_if_invalid=(F[Config].port,),
            ),
            Config,
        )

        assert result.port == 3000
        assert result.inner.port == 9090

    def test_skip_nested_field_not_root(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"port": 8080, "inner": {"port": "abc"}}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"inner": {"port": 9090}}')

        @dataclass
        class Inner:
            port: int

        @dataclass
        class Config:
            port: int
            inner: Inner

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(
                        file_=source1,
                        skip_if_invalid=(F[Config].inner.port,),
                    ),
                    LoadMetadata(file_=source2),
                ),
            ),
            Config,
        )

        assert result.port == 8080
        assert result.inner.port == 9090

    def test_skip_both_root_and_nested(self, tmp_path: Path):
        source1 = tmp_path / "s1.json"
        source1.write_text('{"port": "abc", "inner": {"port": "def"}}')

        source2 = tmp_path / "s2.json"
        source2.write_text('{"port": 8080, "inner": {"port": 9090}}')

        @dataclass
        class Inner:
            port: int

        @dataclass
        class Config:
            port: int
            inner: Inner

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(
                        file_=source1,
                        skip_if_invalid=(F[Config].port, F[Config].inner.port),
                    ),
                    LoadMetadata(file_=source2),
                ),
            ),
            Config,
        )

        assert result.port == 8080
        assert result.inner.port == 9090



"""Callable merge — custom merge function for a field."""

from dataclasses import dataclass
from pathlib import Path
from typing import Any

from dature import F, LoadMetadata, MergeMetadata, MergeRule, MergeStrategy, load

SHARED_DIR = Path(__file__).parents[2] / "shared"


@dataclass
class Config:
    host: str
    port: int
    tags: list[str]


def merge_tags(values: list[Any]) -> list[str]:
    return sorted({v for lst in values for v in lst})


config = load(
    MergeMetadata(
        sources=(
            LoadMetadata(file_=SHARED_DIR / "common_defaults.yaml"),
            LoadMetadata(file_=SHARED_DIR / "common_overrides.yaml"),
        ),
        strategy=MergeStrategy.LAST_WINS,
        field_merges=(MergeRule(F[Config].tags, merge_tags),),
    ),
    Config,
)

assert config.host == "production.example.com"
assert config.port == 8080
assert config.tags == ["api", "default", "web"]



"""Tests for base loader functions (used only in BaseLoader defaults)."""

from datetime import timedelta
from urllib.parse import urlparse

import pytest

from dature.fields.byte_size import ByteSize
from dature.fields.payment_card import PaymentCardNumber
from dature.fields.secret_str import SecretStr
from dature.sources_loader.loaders.base import (
    base64url_bytes_from_string,
    base64url_str_from_string,
    byte_size_from_string,
    bytes_from_string,
    complex_from_string,
    payment_card_number_from_string,
    secret_str_from_string,
    timedelta_from_string,
    url_from_string,
)


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("hello", b"hello"),
    ],
)
def test_bytes_from_string(input_value, expected):
    assert bytes_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("1+2j", 1 + 2j),
    ],
)
def test_complex_from_string(input_value, expected):
    assert complex_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("2:30", timedelta(hours=2, minutes=30)),
        ("2:30:00", timedelta(hours=2, minutes=30)),
        ("0:00:01", timedelta(seconds=1)),
        ("0:45:00", timedelta(minutes=45)),
        ("2:03:04.500000", timedelta(hours=2, minutes=3, seconds=4, microseconds=500000)),
        ("1 day, 2:30:00", timedelta(days=1, hours=2, minutes=30)),
        ("2 days, 0:00:00", timedelta(days=2)),
        ("1 day, 2:03:04.500000", timedelta(days=1, hours=2, minutes=3, seconds=4, microseconds=500000)),
        ("2 days 1:02:03.500000", timedelta(days=2, hours=1, minutes=2, seconds=3, microseconds=500000)),
        ("1 day 2:03:04.500000", timedelta(days=1, hours=2, minutes=3, seconds=4, microseconds=500000)),
        ("2 days 1:02:03", timedelta(days=2, hours=1, minutes=2, seconds=3)),
        ("1 day 2:30:00", timedelta(days=1, hours=2, minutes=30)),
        ("3 days", timedelta(days=3)),
        ("1 day", timedelta(days=1)),
        ("-1 day, 23:59:59", timedelta(days=-1, hours=23, minutes=59, seconds=59)),
        ("-2 days, 23:59:59", timedelta(days=-2, hours=23, minutes=59, seconds=59)),
        ("-1 day 23:59:59", timedelta(days=-1, hours=23, minutes=59, seconds=59)),
        ("-2 days 1:02:03", timedelta(days=-2, hours=1, minutes=2, seconds=3)),
        ("-2:30", timedelta(hours=-2, minutes=-30)),
        ("-0:00:01", timedelta(seconds=-1)),
        ("-2:03:04.500000", timedelta(hours=-2, minutes=-3, seconds=-4, microseconds=-500000)),
        ("1 week", timedelta(weeks=1)),
        ("2 weeks", timedelta(weeks=2)),
        ("-1 week", timedelta(weeks=-1)),
        ("-2 weeks", timedelta(weeks=-2)),
        ("1 week, 2:30:00", timedelta(weeks=1, hours=2, minutes=30)),
        ("2 weeks 1:02:03", timedelta(weeks=2, hours=1, minutes=2, seconds=3)),
        ("1 week 1 day", timedelta(weeks=1, days=1)),
        ("2 weeks, 3 days", timedelta(weeks=2, days=3)),
        ("1 week, 1 day, 2:30:00", timedelta(weeks=1, days=1, hours=2, minutes=30)),
        ("2 weeks 3 days 1:02:03", timedelta(weeks=2, days=3, hours=1, minutes=2, seconds=3)),
        ("-1 week, 1 day, 23:59:59", timedelta(weeks=-1, days=1, hours=23, minutes=59, seconds=59)),
        ("-2 weeks 1:02:03", timedelta(weeks=-2, hours=1, minutes=2, seconds=3)),
    ],
)
def test_timedelta_from_string(input_value: str, expected: timedelta):
    assert timedelta_from_string(input_value) == expected


def test_timedelta_from_string_invalid():
    with pytest.raises(ValueError, match="Invalid timedelta format"):
        timedelta_from_string("not a timedelta")


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (
            "https://example.com/path?query=value#fragment",
            urlparse("https://example.com/path?query=value#fragment"),
        ),
        (
            "http://localhost:8080",
            urlparse("http://localhost:8080"),
        ),
        (
            "ftp://files.example.com/data.csv",
            urlparse("ftp://files.example.com/data.csv"),
        ),
    ],
)
def test_url_from_string(input_value: str, expected):
    assert url_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("SGVsbG8gV29ybGQ=", b"Hello World"),
        ("", b""),
        ("YWJj", b"abc"),
        ("-__-", b"\xfb\xff\xfe"),
    ],
)
def test_base64url_bytes_from_string(input_value: str, expected: bytes):
    assert base64url_bytes_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("SGVsbG8=", "Hello"),
        ("", ""),
        ("0L/RgNC40LLQtdGC", "\u043f\u0440\u0438\u0432\u0435\u0442"),
    ],
)
def test_base64url_str_from_string(input_value: str, expected: str):
    assert base64url_str_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("mysecret", SecretStr("mysecret")),
        ("", SecretStr("")),
    ],
)
def test_secret_str_from_string(input_value: str, expected: SecretStr):
    assert secret_str_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected_brand"),
    [
        ("4111111111111111", "Visa"),
        ("5500000000000004", "Mastercard"),
    ],
)
def test_payment_card_number_from_string(input_value: str, expected_brand: str):
    result = payment_card_number_from_string(input_value)
    assert isinstance(result, PaymentCardNumber)
    assert result.brand == expected_brand


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("1.5 GB", ByteSize(1_500_000_000)),
        (1024, ByteSize(1024)),
    ],
)
def test_byte_size_from_string(input_value: str | int, expected: ByteSize):
    assert byte_size_from_string(input_value) == expected



"""PaymentCardNumber — validates via Luhn and detects brand."""

from dataclasses import dataclass

from dature.fields.payment_card import PaymentCardNumber


@dataclass
class Config:
    card: PaymentCardNumber


config = Config(card=PaymentCardNumber("4111111111111111"))

assert str(config.card) == "************1111"
assert repr(config.card) == "PaymentCardNumber('************1111')"
assert config.card.masked == "************1111"
assert config.card.brand == "Visa"
assert config.card.get_raw_number() == "4111111111111111"



"""Report on error — get_load_report() from the dataclass type after a failed load."""

from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, MergeMetadata, get_load_report, load
from dature.errors.exceptions import DatureConfigError

SOURCES_DIR = Path(__file__).parent / "sources"
SHARED_DIR = Path(__file__).parents[2] / "shared"


@dataclass
class Config:
    host: str
    port: int
    tags: list[str]


try:
    config = load(
        MergeMetadata(
            sources=(
                LoadMetadata(file_=SHARED_DIR / "common_overrides.yaml"),
                LoadMetadata(file_=SOURCES_DIR / "advanced_debug_error_defaults.yaml"),
            ),
        ),
        Config,
        debug=True,
    )
except DatureConfigError:
    report = get_load_report(Config)
    assert report is not None

    assert report.dataclass_name == "Config"
    assert report.strategy == "last_wins"
    assert report.merged_data == {
        "host": "localhost",
        "port": "not_a_number",
        "tags": ["default"],
    }

    assert len(report.sources) == 2

    assert report.sources[0].index == 0
    assert report.sources[0].loader_type == "yaml1.2"
    assert "overrides" in str(report.sources[0].file_path)
    assert report.sources[0].raw_data == {
        "host": "production.example.com",
        "port": 8080,
        "tags": ["web", "api"],
    }

    assert report.sources[1].index == 1
    assert report.sources[1].loader_type == "yaml1.2"
    assert "advanced_debug_error_defaults" in str(report.sources[1].file_path)
    assert report.sources[1].raw_data == {
        "host": "localhost",
        "port": "not_a_number",
        "tags": ["default"],
    }

    assert len(report.field_origins) == 3
    for origin in report.field_origins:
        assert origin.source_index == 1
        assert "advanced_debug_error_defaults" in str(origin.source_file)



"""Strategy is only a priority — if only one form exists, it is always used."""

import os
from dataclasses import dataclass

from dature import LoadMetadata, load
from dature.sources_loader.env_ import EnvLoader

# Only JSON form, no flat keys
os.environ["APP__DATABASE"] = '{"host": "json-host", "port": "5432"}'

# Make sure no flat keys interfere
os.environ.pop("APP__DATABASE__HOST", None)
os.environ.pop("APP__DATABASE__PORT", None)


@dataclass
class Database:
    host: str
    port: int


@dataclass
class Config:
    database: Database


# Even with strategy="flat", JSON is parsed because there are no flat keys
config = load(
    LoadMetadata(loader=EnvLoader, prefix="APP__", nested_resolve_strategy="flat"),
    Config,
)

assert config.database.host == "json-host"
assert config.database.port == 5432



"""Per-field nested_resolve overrides global nested_resolve_strategy."""

import os
from dataclasses import dataclass

from dature import F, LoadMetadata, load
from dature.sources_loader.env_ import EnvLoader

os.environ["APP__DATABASE"] = '{"host": "json-host", "port": "5432"}'
os.environ["APP__DATABASE__HOST"] = "flat-host"
os.environ["APP__DATABASE__PORT"] = "3306"
os.environ["APP__CACHE"] = '{"host": "json-cache", "ttl": "60"}'
os.environ["APP__CACHE__HOST"] = "flat-cache"
os.environ["APP__CACHE__TTL"] = "120"


@dataclass
class Database:
    host: str
    port: int


@dataclass
class Cache:
    host: str
    ttl: int


@dataclass
class Config:
    database: Database
    cache: Cache


# Global: "flat" for everything, but database overridden to "json"
config = load(
    LoadMetadata(
        loader=EnvLoader,
        prefix="APP__",
        nested_resolve_strategy="flat",
        nested_resolve={"json": (F[Config].database,)},
    ),
    Config,
)

assert config.database.host == "json-host"  # per-field override wins
assert config.database.port == 5432
assert config.cache.host == "flat-cache"  # global strategy
assert config.cache.ttl == 120



from typing import Annotated, Any, get_args, get_origin

from adaptix import P, validator
from adaptix.provider import Provider

from dature.expansion.alias_provider import resolve_nested_owner
from dature.field_path import FieldPath
from dature.protocols import DataclassInstance, ValidatorProtocol
from dature.types import FieldValidators


def extract_validators_from_type(field_type: Any) -> list[ValidatorProtocol]:  # noqa: ANN401
    validators: list[ValidatorProtocol] = []

    if get_origin(field_type) is not Annotated:
        return validators

    args = get_args(field_type)

    validators.extend(arg for arg in args[1:] if hasattr(arg, "__dataclass_fields__"))

    return validators


def create_validator_providers(
    dataclass_: type,
    field_name: str,
    validators: list[ValidatorProtocol],
) -> list[Provider]:
    providers = []

    for v in validators:
        func = v.get_validator_func()
        error = v.get_error_message()
        provider = validator(
            P[dataclass_][field_name],
            func,
            error,
        )
        providers.append(provider)

    return providers


def _normalize_validators(
    value: ValidatorProtocol | tuple[ValidatorProtocol, ...],
) -> list[ValidatorProtocol]:
    if isinstance(value, tuple):
        return list(value)
    return [value]


def create_metadata_validator_providers(
    field_validators: FieldValidators,
) -> list[Provider]:
    providers: list[Provider] = []

    for field_path_key, validators_value in field_validators.items():
        if not isinstance(field_path_key, FieldPath):
            msg = f"validators key must be a FieldPath, got {type(field_path_key).__name__}"
            raise TypeError(msg)

        if len(field_path_key.parts) == 0:
            msg = "FieldPath must contain at least one field name"
            raise ValueError(msg)

        normalized = _normalize_validators(validators_value)

        if len(field_path_key.parts) > 1:
            if isinstance(field_path_key.owner, str):
                msg = (
                    f"Nested FieldPath with string owner '{field_path_key.owner}' "
                    f"is not supported — cannot resolve intermediate types"
                )
                raise TypeError(msg)
            owner: type[DataclassInstance] = resolve_nested_owner(field_path_key.owner, field_path_key.parts[:-1])
        else:
            if isinstance(field_path_key.owner, str):
                msg = (
                    f"FieldPath with string owner '{field_path_key.owner}' "
                    f"is not supported for validators — use a type reference"
                )
                raise TypeError(msg)
            owner = field_path_key.owner

        field_name = field_path_key.parts[-1]
        field_providers = create_validator_providers(owner, field_name, normalized)
        providers.extend(field_providers)

    return providers


def create_root_validator_providers(
    dataclass_: type,
    root_validators: tuple[ValidatorProtocol, ...],
) -> list[Provider]:
    providers = []

    for root_validator in root_validators:
        provider = validator(
            P[dataclass_],
            root_validator.get_validator_func(),
            root_validator.get_error_message(),
        )
        providers.append(provider)

    return providers



import json
from collections.abc import Callable
from dataclasses import dataclass
from json.decoder import JSONArray, JSONObject, scanstring  # type: ignore[attr-defined]
from json.scanner import py_make_scanner  # type: ignore[attr-defined]
from typing import TYPE_CHECKING

from dature.errors.exceptions import LineRange
from dature.path_finders.base import PathFinder

if TYPE_CHECKING:
    from dature.types import JSONValue

    _ScanOnce = Callable[[str, int], tuple["JSONValue", int]]


@dataclass(frozen=True, slots=True)
class ExtractedKey:
    key: str | None
    position: int


class JsonPathFinder(PathFinder):
    def __init__(self, content: str) -> None:
        self._line_map = _build_json_line_map(content)

    def find_line_range(self, target_path: list[str]) -> LineRange | None:
        return self._line_map.get(tuple(target_path))


def _build_json_line_map(content: str) -> dict[tuple[str, ...], LineRange]:
    line_map: dict[tuple[str, ...], LineRange] = {}
    path_stack: list[str] = []

    decoder = json.JSONDecoder()

    def _char_to_line(idx: int) -> int:
        return content.count("\n", 0, idx) + 1

    def _wrapping_parse_object(
        s_and_end: tuple[str, int],
        strict: bool,  # noqa: FBT001
        scan_once: "_ScanOnce",
        object_hook: Callable[["JSONValue"], "JSONValue"],
        object_pairs_hook: Callable[["JSONValue"], "JSONValue"],
        memo: dict[str, str] | None = None,
    ) -> tuple["JSONValue", int]:
        def tracking_scan_once(s: str, idx: int) -> tuple["JSONValue", int]:
            extracted = _extract_key_before_value(s, idx)
            if extracted.key is not None:
                path_stack.append(extracted.key)

            obj, end = scan_once(s, idx)

            if extracted.key is not None:
                line_map[tuple(path_stack)] = LineRange(
                    start=_char_to_line(extracted.position),
                    end=_char_to_line(end - 1),
                )
                path_stack.pop()

            return obj, end

        return JSONObject(  # type: ignore[no-any-return]
            s_and_end,
            strict,
            tracking_scan_once,
            object_hook,
            object_pairs_hook,
            memo,
        )

    def _wrapping_parse_array(
        s_and_end: tuple[str, int],
        scan_once: "_ScanOnce",
    ) -> tuple[list["JSONValue"], int]:
        idx = 0

        def tracking_scan_once(s: str, pos: int) -> tuple["JSONValue", int]:
            nonlocal idx
            path_stack.append(str(idx))
            obj, end = scan_once(s, pos)
            path_stack.pop()
            idx += 1
            return obj, end

        return JSONArray(s_and_end, tracking_scan_once)  # type: ignore[no-any-return]

    decoder.parse_object = _wrapping_parse_object  # type: ignore[attr-defined]
    decoder.parse_array = _wrapping_parse_array  # type: ignore[attr-defined]
    decoder.scan_once = py_make_scanner(decoder)  # type: ignore[attr-defined]
    decoder.decode(content)
    return line_map


def _extract_key_before_value(s: str, val_start: int) -> ExtractedKey:
    """Находит ключ JSON-объекта, отступая назад от позиции начала значения."""
    pos = val_start - 1
    while pos >= 0 and s[pos] in " \t\n\r:":
        pos -= 1

    if pos < 0 or s[pos] != '"':
        return ExtractedKey(key=None, position=val_start)

    pos -= 1
    while pos >= 0:
        if s[pos] == '"':
            num_backslashes = 0
            check = pos - 1
            while check >= 0 and s[check] == "\\":
                num_backslashes += 1
                check -= 1
            if num_backslashes % 2 == 0:
                key, _ = scanstring(s, pos + 1, True)  # noqa: FBT003
                return ExtractedKey(key=key, position=pos)
        pos -= 1

    return ExtractedKey(key=None, position=val_start)



from json5.model import Identifier, JSONArray, JSONObject, String, Value
from json5.parser import parse_source

from dature.errors.exceptions import LineRange
from dature.path_finders.base import PathFinder


class Json5PathFinder(PathFinder):
    def __init__(self, content: str) -> None:
        self._line_map = _build_json5_line_map(content)

    def find_line_range(self, target_path: list[str]) -> LineRange | None:
        return self._line_map.get(tuple(target_path))


def _build_json5_line_map(content: str) -> dict[tuple[str, ...], LineRange]:
    model = parse_source(content)
    line_map: dict[tuple[str, ...], LineRange] = {}
    _walk(model.value, (), line_map)
    return line_map


def _walk(node: Value, parent_path: tuple[str, ...], line_map: dict[tuple[str, ...], LineRange]) -> None:
    if isinstance(node, JSONObject):
        _walk_object(node, parent_path, line_map)
    elif isinstance(node, JSONArray):
        _walk_array(node, parent_path, line_map)


def _walk_object(
    obj: JSONObject,
    parent_path: tuple[str, ...],
    line_map: dict[tuple[str, ...], LineRange],
) -> None:
    for kvp in obj.key_value_pairs:
        key_node = kvp[0]
        val_node = kvp[1]

        if isinstance(key_node, Identifier):
            key_name = key_node.name
        elif isinstance(key_node, String):
            key_name = key_node.characters
        else:
            continue

        current_path = (*parent_path, key_name)

        start = key_node.lineno
        end = val_node.end_lineno
        if start is None or end is None:
            continue

        line_map[current_path] = LineRange(start=start, end=end)
        _walk(val_node, current_path, line_map)


def _walk_array(
    arr: JSONArray,
    parent_path: tuple[str, ...],
    line_map: dict[tuple[str, ...], LineRange],
) -> None:
    for index, val_node in enumerate(arr.values):
        current_path = (*parent_path, str(index))
        _walk(val_node, current_path, line_map)



"""ByteSize — parses human-readable sizes."""

from dataclasses import dataclass

from dature.fields.byte_size import ByteSize


@dataclass
class Config:
    max_upload: ByteSize


config = Config(max_upload=ByteSize("1.5 GB"))

assert int(config.max_upload) == 1500000000
assert str(config.max_upload) == "1.4GiB"
assert repr(config.max_upload) == "ByteSize(1500000000)"
assert config.max_upload.human_readable() == "1.4GiB"
assert config.max_upload.human_readable(decimal=True) == "1.5GB"
assert ByteSize(1024) > ByteSize(512)



import io
import os
from collections.abc import Iterable
from pathlib import Path
from typing import ClassVar, cast

from dature.errors.exceptions import LineRange, SourceLocation
from dature.sources_loader.flat_key import FlatKeyLoader
from dature.types import (
    BINARY_IO_TYPES,
    TEXT_IO_TYPES,
    FileOrStream,
    JSONValue,
    NestedConflict,
    NestedConflicts,
)


class EnvLoader(FlatKeyLoader):
    display_name = "env"
    display_label: ClassVar[str] = "ENV"

    def _load(self, _: FileOrStream) -> JSONValue:
        return cast("JSONValue", os.environ)

    @classmethod
    def resolve_location(
        cls,
        field_path: list[str],
        file_path: Path | None,  # noqa: ARG003
        file_content: str | None,  # noqa: ARG003
        prefix: str | None,
        split_symbols: str,
        nested_conflict: NestedConflict | None,
    ) -> list[SourceLocation]:
        var_name = cls._resolve_var_name(field_path, prefix, split_symbols, nested_conflict)
        env_var_value: str | None = None
        if nested_conflict is not None:
            json_var = cls._resolve_var_name(field_path[:1], prefix, split_symbols, None)
            if nested_conflict.used_var == json_var:
                env_var_value = nested_conflict.json_raw_value
        return [
            SourceLocation(
                display_label=cls.display_label,
                file_path=None,
                line_range=None,
                line_content=None,
                env_var_name=var_name,
                env_var_value=env_var_value,
            ),
        ]

    def _pre_process_row(
        self,
        key: str,
        value: str,
        result: dict[str, JSONValue],
        conflicts: NestedConflicts,
    ) -> None:
        if self._prefix and not key.startswith(self._prefix):
            return

        processed_key = key[len(self._prefix) :] if self._prefix else key
        processed_key = processed_key.lower()

        parts = processed_key.split(self._split_symbols)
        self._process_key_value(parts=parts, value=value, result=result, conflicts=conflicts)


class EnvFileLoader(EnvLoader):
    display_name = "envfile"
    display_label: ClassVar[str] = "ENV FILE"

    @classmethod
    def resolve_location(
        cls,
        field_path: list[str],
        file_path: Path | None,
        file_content: str | None,
        prefix: str | None,
        split_symbols: str,
        nested_conflict: NestedConflict | None,
    ) -> list[SourceLocation]:
        var_name = cls._resolve_var_name(field_path, prefix, split_symbols, nested_conflict)
        line_range: LineRange | None = None
        line_content: list[str] | None = None
        if file_content is not None:
            line_range, line_content = _find_env_line(file_content, var_name)
        return [
            SourceLocation(
                display_label=cls.display_label,
                file_path=file_path,
                line_range=line_range,
                line_content=line_content,
                env_var_name=var_name,
            ),
        ]

    def _load(self, path: FileOrStream) -> JSONValue:
        """Parse .env file into a flat key=value dict (before nesting/expand/parse)."""
        raw_pairs: dict[str, JSONValue] = {}

        if isinstance(path, TEXT_IO_TYPES):
            self._collect_pairs(path, raw_pairs)
        elif isinstance(path, BINARY_IO_TYPES):
            wrapper = io.TextIOWrapper(cast("io.BufferedReader", path))
            self._collect_pairs(wrapper, raw_pairs)
        else:
            with path.open() as f:
                self._collect_pairs(f, raw_pairs)

        return raw_pairs

    @staticmethod
    def _collect_pairs(
        lines: Iterable[str],
        pairs: dict[str, JSONValue],
    ) -> None:
        for raw_line in lines:
            if not (line := raw_line.strip()) or line.startswith("#"):
                continue

            if "=" not in line:
                continue

            key, value = line.split("=", 1)
            key = key.strip()
            value = value.strip()
            _min_quoted_len = 2
            if len(value) >= _min_quoted_len and value[0] == value[-1] and value[0] in ('"', "'"):
                value = value[1:-1]
            pairs[key] = value


def _find_env_line(content: str, var_name: str) -> tuple[LineRange | None, list[str] | None]:
    for i, line in enumerate(content.splitlines(), 1):
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        if "=" not in stripped:
            continue
        key = stripped.split("=", 1)[0].strip()
        if key == var_name:
            return LineRange(start=i, end=i), [stripped]
    return None, None



"""Per-merge type_loaders — set type_loaders on MergeMetadata for multi-source loads."""

from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, MergeMetadata, TypeLoader, load

SOURCES_DIR = Path(__file__).parent / "sources"


@dataclass(frozen=True, slots=True)
class Rgb:
    r: int
    g: int
    b: int


def rgb_from_string(value: str) -> Rgb:
    parts = value.split(",")
    return Rgb(r=int(parts[0]), g=int(parts[1]), b=int(parts[2]))


@dataclass
class AppConfig:
    name: str
    color: Rgb


config = load(
    MergeMetadata(
        sources=(
            LoadMetadata(file_=SOURCES_DIR / "custom_type_common.yaml"),
            LoadMetadata(file_=SOURCES_DIR / "custom_type_merge_override.yaml"),
        ),
        type_loaders=(TypeLoader(type_=Rgb, func=rgb_from_string),),
    ),
    AppConfig,
)

assert config == AppConfig(name="my-app", color=Rgb(r=100, g=200, b=50))



"""field_mapping — explicit field renaming with F objects."""

from dataclasses import dataclass
from pathlib import Path

from dature import F, LoadMetadata, load

SOURCES_DIR = Path(__file__).parent / "sources"


@dataclass
class DbConfig:
    database_url: str
    secret_key: str
    pool_size: int


config = load(
    LoadMetadata(
        file_=SOURCES_DIR / "naming_field_mapping.yaml",
        field_mapping={
            F[DbConfig].database_url: "db_url",
            F[DbConfig].secret_key: "key",
            F[DbConfig].pool_size: "pool",
        },
    ),
    DbConfig,
)

assert config.database_url == "postgresql://localhost:5432/mydb"
assert config.secret_key == "my-secret-key"
assert config.pool_size == 10



from dataclasses import dataclass, field
from pathlib import Path

import pytest

from dature import LoadMetadata, load


class TestPostInitValidationFunctionMode:
    def test_post_init_success(self, tmp_path: Path):
        @dataclass
        class Config:
            port: int
            host: str

            def __post_init__(self) -> None:
                if self.port < 0 or self.port > 65535:
                    msg = f"Invalid port: {self.port}"
                    raise ValueError(msg)

        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 8080, "host": "localhost"}')

        result = load(LoadMetadata(file_=json_file), Config)

        assert result.port == 8080
        assert result.host == "localhost"

    def test_post_init_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            port: int
            host: str

            def __post_init__(self) -> None:
                if self.port < 0 or self.port > 65535:
                    msg = f"Invalid port: {self.port}"
                    raise ValueError(msg)

        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 99999, "host": "localhost"}')

        with pytest.raises(ValueError, match="Invalid port: 99999"):
            load(LoadMetadata(file_=json_file), Config)

    def test_post_init_cross_field_validation(self, tmp_path: Path):
        @dataclass
        class Config:
            min_value: int
            max_value: int

            def __post_init__(self) -> None:
                if self.min_value >= self.max_value:
                    msg = f"min_value ({self.min_value}) must be less than max_value ({self.max_value})"
                    raise ValueError(msg)

        json_file = tmp_path / "config.json"
        json_file.write_text('{"min_value": 100, "max_value": 10}')

        with pytest.raises(ValueError, match=r"min_value \(100\) must be less than max_value \(10\)"):
            load(LoadMetadata(file_=json_file), Config)

    def test_post_init_cross_field_success(self, tmp_path: Path):
        @dataclass
        class Config:
            min_value: int
            max_value: int

            def __post_init__(self) -> None:
                if self.min_value >= self.max_value:
                    msg = f"min_value ({self.min_value}) must be less than max_value ({self.max_value})"
                    raise ValueError(msg)

        json_file = tmp_path / "config.json"
        json_file.write_text('{"min_value": 1, "max_value": 100}')

        result = load(LoadMetadata(file_=json_file), Config)

        assert result.min_value == 1
        assert result.max_value == 100


class TestPostInitValidationDecoratorMode:
    def test_post_init_success(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 8080, "host": "localhost"}')

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            port: int
            host: str

            def __post_init__(self) -> None:
                if self.port < 0 or self.port > 65535:
                    msg = f"Invalid port: {self.port}"
                    raise ValueError(msg)

        config = Config()

        assert config.port == 8080
        assert config.host == "localhost"

    def test_post_init_failure_from_file(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 99999, "host": "localhost"}')

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            port: int
            host: str

            def __post_init__(self) -> None:
                if self.port < 0 or self.port > 65535:
                    msg = f"Invalid port: {self.port}"
                    raise ValueError(msg)

        with pytest.raises(ValueError, match="Invalid port: 99999"):
            Config()

    def test_post_init_failure_from_override(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 8080, "host": "localhost"}')

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            port: int
            host: str

            def __post_init__(self) -> None:
                if self.port < 0 or self.port > 65535:
                    msg = f"Invalid port: {self.port}"
                    raise ValueError(msg)

        with pytest.raises(ValueError, match="Invalid port: -1"):
            Config(port=-1)

    def test_post_init_cross_field(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"min_value": 50, "max_value": 10}')

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            min_value: int
            max_value: int

            def __post_init__(self) -> None:
                if self.min_value >= self.max_value:
                    msg = "min must be less than max"
                    raise ValueError(msg)

        with pytest.raises(ValueError, match="min must be less than max"):
            Config()


class TestPostInitComputedFields:
    def test_computed_field_via_post_init(self, tmp_path: Path):
        @dataclass
        class Config:
            host: str
            port: int
            base_url: str = field(init=False)

            def __post_init__(self) -> None:
                self.base_url = f"http://{self.host}:{self.port}"

        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": 8080}')

        result = load(LoadMetadata(file_=json_file), Config)

        assert result.base_url == "http://localhost:8080"

    def test_computed_field_via_post_init_decorator(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "example.com", "port": 443}')

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            host: str
            port: int
            base_url: str = ""

            def __post_init__(self) -> None:
                self.base_url = f"https://{self.host}:{self.port}"

        config = Config()

        assert config.base_url == "https://example.com:443"


class TestPropertyValidation:
    def test_property_computed_value(self, tmp_path: Path):
        @dataclass
        class Config:
            host: str
            port: int

            @property
            def address(self) -> str:
                return f"{self.host}:{self.port}"

        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": 8080}')

        result = load(LoadMetadata(file_=json_file), Config)

        assert result.address == "localhost:8080"

    def test_property_computed_value_decorator(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": 3000}')

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            host: str
            port: int

            @property
            def address(self) -> str:
                return f"{self.host}:{self.port}"

        config = Config()

        assert config.address == "localhost:3000"

    def test_property_with_validation_logic(self, tmp_path: Path):
        @dataclass
        class Config:
            _email: str

            @property
            def email(self) -> str:
                return self._email.lower().strip()

        json_file = tmp_path / "config.json"
        json_file.write_text('{"_email": "  Admin@Example.COM  "}')

        result = load(LoadMetadata(file_=json_file), Config)

        assert result.email == "admin@example.com"



import math
import re
from dataclasses import fields

from examples.all_types_dataclass import AllPythonTypesCompact


def assert_all_types_equal(result: AllPythonTypesCompact, expected: AllPythonTypesCompact) -> None:
    for field in fields(result):
        result_val = getattr(result, field.name)
        expected_val = getattr(expected, field.name)
        if (
            isinstance(result_val, float)
            and isinstance(expected_val, float)
            and math.isnan(result_val)
            and math.isnan(expected_val)
        ):
            continue
        if isinstance(result_val, re.Pattern) and isinstance(expected_val, re.Pattern):
            assert result_val.pattern == expected_val.pattern, (
                f"{field.name}: pattern {result_val.pattern!r} != {expected_val.pattern!r}"
            )
            assert result_val.flags == expected_val.flags, (
                f"{field.name}: flags {result_val.flags!r} != {expected_val.flags!r}"
            )
            continue
        assert result_val == expected_val, f"{field.name}: {result_val!r} != {expected_val!r}"



"""nested_resolve_strategy with Docker secrets source."""

from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory

from dature import LoadMetadata, load
from dature.sources_loader.docker_secrets import DockerSecretsLoader


@dataclass
class Database:
    host: str
    port: int


@dataclass
class Config:
    database: Database


with TemporaryDirectory() as secrets_dir:
    secrets_path = Path(secrets_dir)
    (secrets_path / "database").write_text('{"host": "json-host", "port": "5432"}')
    (secrets_path / "database__host").write_text("flat-host")
    (secrets_path / "database__port").write_text("3306")

    config = load(
        LoadMetadata(
            file_=secrets_path,
            loader=DockerSecretsLoader,
            nested_resolve_strategy="json",
        ),
        Config,
    )

    assert config.database.host == "json-host"
    assert config.database.port == 5432



from pathlib import Path
from typing import TYPE_CHECKING, Any

from dature.sources_loader.docker_secrets import DockerSecretsLoader
from dature.sources_loader.env_ import EnvFileLoader, EnvLoader
from dature.sources_loader.ini_ import IniLoader
from dature.sources_loader.json_ import JsonLoader
from dature.types import FILE_LIKE_TYPES, ExpandEnvVarsMode, NestedResolve, NestedResolveStrategy

if TYPE_CHECKING:
    from dature.metadata import LoadMetadata, TypeLoader
    from dature.protocols import LoaderProtocol
    from dature.types import FileLike, FilePath

SUPPORTED_EXTENSIONS = (".cfg", ".env", ".ini", ".json", ".json5", ".toml", ".yaml", ".yml")

_EXTRA_BY_EXTENSION: dict[str, str] = {
    ".toml": "toml",
    ".yaml": "yaml",
    ".yml": "yaml",
    ".json5": "json5",
}


def _resolve_by_extension(extension: str) -> "type[LoaderProtocol]":
    try:
        return _resolve_by_extension_inner(extension)
    except ImportError:
        extra = _EXTRA_BY_EXTENSION.get(extension)
        if extra is None:
            raise
        msg = f"To use '{extension}' files, install the '{extra}' extra: pip install dature[{extra}]"
        raise ImportError(msg) from None


def _resolve_by_extension_inner(extension: str) -> "type[LoaderProtocol]":
    match extension:
        case ".json":
            return JsonLoader
        case ".toml":
            from dature.sources_loader.toml_ import Toml11Loader  # noqa: PLC0415

            return Toml11Loader
        case ".ini" | ".cfg":
            return IniLoader
        case ".env":
            return EnvFileLoader
        case ".yaml" | ".yml":
            from dature.sources_loader.yaml_ import Yaml12Loader  # noqa: PLC0415

            return Yaml12Loader
        case ".json5":
            from dature.sources_loader.json5_ import Json5Loader  # noqa: PLC0415

            return Json5Loader
        case _:
            supported = ", ".join(SUPPORTED_EXTENSIONS)
            msg = (
                f"Cannot determine loader type for extension '{extension}'. "
                f"Please specify loader explicitly or use a supported extension: {supported}"
            )
            raise ValueError(msg)


def resolve_loader_class(
    loader: "type[LoaderProtocol] | None",
    file_: "FileLike | FilePath | None",
) -> "type[LoaderProtocol]":
    if loader is not None:
        if file_ is not None and not isinstance(file_, FILE_LIKE_TYPES) and loader is EnvLoader:
            msg = (
                "EnvLoader reads from environment variables and does not use files. "
                "Remove file_ or use a file-based loader instead (e.g. EnvFileLoader)."
            )
            raise ValueError(msg)
        if isinstance(file_, FILE_LIKE_TYPES) and loader in (EnvLoader, DockerSecretsLoader):
            msg = (
                f"{loader.__name__} does not support file-like objects. "
                "Use a file-based loader (e.g. JsonLoader, TomlLoader) with file-like objects."
            )
            raise ValueError(msg)
        return loader

    if isinstance(file_, FILE_LIKE_TYPES):
        msg = (
            "Cannot determine loader type for a file-like object. "
            "Please specify loader explicitly (e.g. loader=JsonLoader)."
        )
        raise TypeError(msg)

    if file_ is None:
        return EnvLoader

    # file-like objects are handled above; here file_ is str | Path
    file_path = Path(file_)

    if file_path.is_dir():
        return DockerSecretsLoader

    if file_path.name.startswith(".env"):
        return EnvFileLoader

    return _resolve_by_extension(file_path.suffix.lower())


def resolve_loader(
    metadata: "LoadMetadata",
    *,
    expand_env_vars: ExpandEnvVarsMode | None = None,
    type_loaders: "tuple[TypeLoader, ...]" = (),
    nested_resolve_strategy: NestedResolveStrategy = "flat",
    nested_resolve: NestedResolve | None = None,
) -> "LoaderProtocol":
    loader_class = resolve_loader_class(metadata.loader, metadata.file_)

    resolved_expand = expand_env_vars or metadata.expand_env_vars or "default"

    kwargs: dict[str, Any] = {
        "prefix": metadata.prefix,
        "name_style": metadata.name_style,
        "field_mapping": metadata.field_mapping,
        "root_validators": metadata.root_validators,
        "validators": metadata.validators,
        "expand_env_vars": resolved_expand,
        "type_loaders": type_loaders,
    }

    if issubclass(loader_class, (EnvLoader, DockerSecretsLoader)):
        kwargs["split_symbols"] = metadata.split_symbols
        resolved_strategy = metadata.nested_resolve_strategy or nested_resolve_strategy
        kwargs["nested_resolve_strategy"] = resolved_strategy
        resolved_resolve = metadata.nested_resolve or nested_resolve
        if resolved_resolve is not None:
            kwargs["nested_resolve"] = resolved_resolve

    return loader_class(**kwargs)
